"""generate openapi docs."""
from pkg_resources import get_distribution
from uwg_schema._openapi import get_openapi
from uwg_schema.model import UWG

import json
import argparse

parser = argparse.ArgumentParser(description='Generate OpenAPI JSON schemas')

parser.add_argument('--version', help='Set the version of the new OpenAPI Schema')

args = parser.parse_args()

if args.version:
    VERSION = args.version.replace('v', '')
else:
    VERSION = '.'.join(get_distribution('uwg_schema').version.split('.')[:3])

info = {
    "description": "",
    "version": VERSION,
    "title": "",
    "contact": {
        "name": "Ladybug Tools",
        "email": "info@ladybug.tools",
        "url": "https://github.com/ladybug-tools/uwg-schema"
    },
    "x-logo": {
        "url": "https://github.com/ladybug-tools/artwork/raw/master/icons_components/dragonfly/png/uwg.png",
        "altText": "UWG logo"
    },
    "license": {
        "name": "MIT",
        "url": "https://github.com/ladybug-tools/uwg-schema/blob/master/LICENSE"
    }
}


# generate Model open api schema
print('Generating UWG Model documentation...')

external_docs = {
    "description": "OpenAPI Specification",
    "url": "./uwg.json"
}

openapi = get_openapi(
    [UWG],
    title='UWG Model Schema',
    description='This is the documentation for UWG model schema.',
    version=VERSION, info=info,
    external_docs=external_docs)
# set the version default key in the UWG schema
openapi['components']['schemas']['UWG']['properties']['version']['default'] = VERSION
with open('./docs/uwg.json', 'w') as out_file:
    json.dump(openapi, out_file, indent=2)



# coding=utf-8
from uwg import Material, Element, Building, BEMDef, SchDef, UWG

import os
import json


def material(directory):
    """Generate Material json."""

    dest_file = os.path.join(directory, 'material.json')
    insul = Material(thermalcond=0.049, volheat=836.8 *
                     265.0, name='insulation')

    with open(dest_file, 'w') as fp:
        json.dump(insul.to_dict(), fp, indent=4)


def element(directory):
    """Generate Element json."""

    dest_file = os.path.join(directory, 'element.json')
    insulation = Material(0.049, 836.8 * 265.0, 'insulation')
    gypsum = Material(0.16, 830.0 * 784.9, 'gypsum')
    wood = Material(0.11, 1210.0 * 544.62, 'wood')
    layer_thickness_lst = [0.01, 0.01, 0.0127]
    material_lst = [wood, insulation, gypsum]
    wall = Element(albedo=0.22, emissivity=0.92, layer_thickness_lst=layer_thickness_lst,
                   material_lst=material_lst, vegcoverage=0, t_init=293,
                   horizontal=False, name='wood_frame_wall')

    with open(dest_file, 'w') as fp:
        json.dump(wall.to_dict(), fp, indent=4)


def building(directory):
    """Generate Building json."""

    dest_file = os.path.join(directory, 'building.json')
    # New Midrise Apartment, 1A
    bldg = Building(
        floor_height=3.0, int_heat_night=1, int_heat_day=1, int_heat_frad=0.1,
        int_heat_flat=0.1, infil=0.171, vent=0.00045, glazing_ratio=0.4, u_value=3.0,
        shgc=0.3, condtype='AIR', cop=3, coolcap=41, heateff=0.8, initial_temp=293)

    with open(dest_file, 'w') as fp:
        json.dump(bldg.to_dict(), fp, indent=4)


def schdef(directory):
    """Generate SchDef json."""

    dest_file = os.path.join(directory, 'schdef.json')
    default_week = [[0.15] * 24] * 3
    schdef = SchDef(elec=default_week, gas=default_week, light=default_week,
                    occ=default_week, cool=default_week, heat=default_week,
                    swh=default_week, q_elec=18.9, q_gas=3.2, q_light=18.9,
                    n_occ=0.12, vent=0.0013, v_swh=0.2846, bldtype='largeoffice',
                    builtera='new')

    with open(dest_file, 'w') as fp:
        json.dump(schdef.to_dict(), fp, indent=4)


def bemdef(directory):
    """Generate BEMDef json."""

    # materials
    insulation = Material(0.049, 836.8 * 265.0, 'insulation')
    gypsum = Material(0.16, 830.0 * 784.9, 'gypsum')
    wood = Material(0.11, 1210.0 * 544.62, 'wood')

    # elements
    wall = Element(0.22, 0.92, [0.01, 0.01, 0.0127], [wood, insulation, gypsum], 0, 293,
                   False, 'wood_frame_wall')
    roof = Element(0.22, 0.92, [0.01, 0.01, 0.0127], [wood, insulation, gypsum], 0, 293,
                   True, 'wood_frame_roof')
    mass = Element(0.2, 0.9, [0.05, 0.05], [
                   wood, wood], 0, 293, True, 'wood_floor')

    # building
    bldg = Building(
        floor_height=3.0, int_heat_night=1, int_heat_day=1, int_heat_frad=0.1,
        int_heat_flat=0.1, infil=0.171, vent=0.00045, glazing_ratio=0.4, u_value=3.0,
        shgc=0.3, condtype='AIR', cop=3, coolcap=41, heateff=0.8, initial_temp=293)
    bemdef = BEMDef(building=bldg, mass=mass, wall=wall, roof=roof, bldtype='largeoffice',
                    builtera='new')

    dest_file = os.path.join(directory, 'bemdef.json')
    with open(dest_file, 'w') as fp:
        json.dump(bemdef.to_dict(), fp, indent=4)


def uwg(directory):
    """Generate UWg json."""
    model = UWG.from_param_args(
        epw_path=None, bldheight=10.0, blddensity=0.5, vertohor=0.5, zone='1A',
        treecover=0.1, grasscover=0.1)

    dest_file = os.path.join(directory, 'uwg.json')
    with open(dest_file, 'w') as fp:
        json.dump(model.to_dict(), fp, indent=4)


def custom_uwg(directory):
    """Generate UWG json with custom reference BEMDef and SchDef objects."""

    # override at 5,2 and add at 18,2

    # SchDef
    default_week = [[0.15] * 24] * 3
    schdef1 = SchDef(elec=default_week, gas=default_week, light=default_week,
                     occ=default_week, cool=default_week, heat=default_week,
                     swh=default_week, q_elec=18.9, q_gas=3.2, q_light=18.9,
                     n_occ=0.12, vent=0.0013, v_swh=0.2846, bldtype='largeoffice',
                     builtera='new')
    default_week = [[0.35] * 24] * 3
    schdef2 = SchDef(elec=default_week, gas=default_week, light=default_week,
                     occ=default_week, cool=default_week, heat=default_week,
                     swh=default_week,  q_elec=18.9, q_gas=3.2, q_light=18.9,
                     n_occ=0.12, vent=0.0013, v_swh=0.2846,
                     bldtype='customhospital', builtera='new')

    # BEMDedf

    # materials
    insulation = Material(0.049, 836.8 * 265.0, 'insulation')
    gypsum = Material(0.16, 830.0 * 784.9, 'gypsum')
    wood = Material(0.11, 1210.0 * 544.62, 'wood')

    # elements
    wall = Element(0.22, 0.92, [0.01, 0.01, 0.0127], [wood, insulation, gypsum], 0, 293,
                   False, 'wood_frame_wall')
    roof = Element(0.22, 0.92, [0.01, 0.01, 0.0127], [wood, insulation, gypsum], 0, 293,
                   True, 'wood_frame_roof')
    mass = Element(0.2, 0.9, [0.05, 0.05], [
                   wood, wood], 0, 293, True, 'wood_floor')

    # building
    bldg = Building(
        floor_height=3.0, int_heat_night=1, int_heat_day=1, int_heat_frad=0.1,
        int_heat_flat=0.1, infil=0.171, vent=0.00045, glazing_ratio=0.4, u_value=3.0,
        shgc=0.3, condtype='AIR', cop=3, coolcap=41, heateff=0.8, initial_temp=293)

    bemdef1 = BEMDef(building=bldg, mass=mass, wall=wall, roof=roof,
                     bldtype='largeoffice', builtera='new')
    bemdef2 = BEMDef(building=bldg, mass=mass, wall=wall, roof=roof,
                     bldtype='customhospital', builtera='new')

    # vectors
    ref_sch_vector = [schdef1, schdef2]
    ref_bem_vector = [bemdef1, bemdef2]
    bld = [('largeoffice', 'new', 0.4),  # overwrite
           ('hospital', 'new', 0.5),
           ('customhospital', 'new', 0.1)]  # extend

    model = UWG.from_param_args(
        epw_path=None, bldheight=10.0, blddensity=0.5, vertohor=0.5, zone='1A',
        treecover=0.1, grasscover=0.1, bld=bld, ref_bem_vector=ref_bem_vector,
        ref_sch_vector=ref_sch_vector)

    dest_file = os.path.join(directory, 'custom_uwg.json')
    with open(dest_file, 'w') as fp:
        json.dump(model.to_dict(include_refDOE=True), fp, indent=4)


if __name__ == '__main__':
    # run all functions within the file
    master_dir = os.path.split(os.path.dirname(__file__))[0]
    sample_directory = os.path.join(master_dir, 'samples')

    material(sample_directory)
    element(sample_directory)
    building(sample_directory)
    schdef(sample_directory)
    bemdef(sample_directory)
    uwg(sample_directory)
    custom_uwg(sample_directory)



from pydantic.utils import get_model
from pydantic.schema import schema, get_flat_models_from_model, get_model_name_map
from typing import Dict, List, Any
import enum

# base open api dictionary for all schemas
_base_open_api = {
    "openapi": "3.0.2",
    "servers": [],
    "info": {},
    "externalDocs": {},
    "tags": [],
    "x-tagGroups": [
        {
            "name": "Models",
            "tags": []
        }
    ],
    "paths": {},
    "components": {"schemas": {}}
}


def get_openapi(
    base_object: List[Any],
    title: str = None,
    version: str = None,
    openapi_version: str = "3.0.2",
    description: str = None,
    info: dict = None,
    external_docs: dict = None
        ) -> Dict:
    """Return UWG Schema as an openapi compatible dictionary."""
    open_api = dict(_base_open_api)

    open_api['openapi'] = openapi_version

    if info:
        open_api['info'] = info

    if title:
        open_api['info']['title'] = title

    if not version:
        raise ValueError(
            'Schema version must be specified as argument or from distribution metadata'
        )

    if version:
        open_api['info']['version'] = version

    if description:
        open_api['info']['description'] = description

    if external_docs:
        open_api['externalDocs'] = external_docs

    schemas = schema(base_object, ref_prefix='#/components/schemas/')['definitions']

    # goes to tags
    tags = []
    # goes to x-tagGroups['tags']
    tag_names = []

    schema_names = list(schemas.keys())
    schema_names.sort()

    for name in schema_names:
        model_name, tag = create_tag(name)
        tag_names.append(model_name)
        tags.append(tag)

        # sort properties order: put required parameters at begining of the list
        s = schemas[name]

        if 'properties' in s:
            properties = s['properties']
        elif 'enum' in s:
            # enum
            continue
        else:
            properties = s['allOf'][1]['properties']

        # make all types readOnly
        try:
            properties['type']['readOnly'] = True
        except KeyError:
            # no type has been set in properties for this object
            typ = {
                'title': 'Type', 'default': f'{name}', 'type': 'string',
                'pattern': f'^{name}$', 'readOnly': True,
            }
            properties['type'] = typ
        # add format to numbers and integers
        # this is helpful for C# generators
        for prop in properties:
            try:
                properties[prop] = set_format(properties[prop])
            except KeyError:
                # referenced object
                if 'anyOf' in properties[prop]:
                    new_any_of = []
                    for item in properties[prop]['anyOf']:
                        new_any_of.append(set_format(item))
                    properties[prop]['anyOf'] = new_any_of
                else:
                    continue

        # sort fields to keep required ones on top
        if 'required' in s:
            required = s['required']
        elif 'allOf' in s:
            try:
                required = s['allOf'][1]['required']
            except KeyError:
                # no required field
                continue
        else:
            continue

        sorted_props = {}
        optional = {}
        for prop, value in properties.items():
            if prop in required:
                sorted_props[prop] = value
            else:
                optional[prop] = value

        sorted_props.update(optional)

        if 'properties' in s:
            s['properties'] = sorted_props
        else:
            s['allOf'][1]['properties'] = sorted_props

    tag_names.sort()
    open_api['tags'] = tags
    open_api['x-tagGroups'][0]['tags'] = tag_names

    open_api['components']['schemas'] = schemas

    return open_api


def create_tag(name):
    """create a viewer tag from a class name."""
    model_name = '%s_model' % name.lower()
    tag = {
        'name': model_name,
        'x-displayName': name,
        'description':
            '<SchemaDefinition schemaRef=\"#/components/schemas/%s\" />\n' % name
    }
    return model_name, tag


def set_format(p):
    """Set format for a property."""
    if '$ref' in p:
        return p
    elif p['type'] == 'number' and 'format' not in p:
        p['format'] = 'double'
    elif p['type'] == 'integer' and 'format' not in p:
        p['format'] = 'int32'
    elif p['type'] == 'array':
        p['items'] = set_format(p['items'])
    return p



from pydantic import Field, validator, root_validator, constr, conlist
from typing import List
from enum import Enum

from ._base import NoExtraBaseModel

WEEK_MATRIX = \
    conlist(conlist(float, min_items=24, max_items=24),
            min_items=3, max_items=3)
REF_BUILTERA = ('pre80', 'pst80', 'new')
REF_BUILTERA_SET = {'pre80', 'pst80', 'new'}


class Material(NoExtraBaseModel):
    """Material class."""

    type: constr(regex='^Material$') = 'Material'

    thermalcond: float = Field(
        ...,
        gt=0,
        description='Number for thermal conductivity [W/(m-K)].'
    )

    volheat: float = Field(
        ...,
        gt=0,
        description='Number for volumetric capacity [J/(m3-K)].'
    )

    name: str = Field(
        ...,
        name='Text string for name of the Material.'
    )


class Element(NoExtraBaseModel):
    """Element object defines constructions."""

    type: constr(regex='^Element$') = 'Element'

    albedo: float = Field(
        ...,
        ge=0,
        le=1,
        description='A value between 0 and 1 for outer surface albedo.'
    )

    emissivity: float = Field(
        ...,
        ge=0,
        le=1,
        description='A value between 0 and 1 for outer surface emissivity.'
    )

    layer_thickness_lst: List[float] = Field(
        ...,
        min_items=1,
        description='List of thickness in meters of each Material in the Element. '
        'The order of thickness should correspond to the order of the Material '
        'objects in material_lst.'
    )

    material_lst: List[Material] = Field(
        ...,
        min_itmes=1,
        description='List of Material objects in the element. The order of Material '
        'objects should correspond to the order of the thickness in layer_thickness_lst.'
    )

    vegcoverage: float = Field(
        ...,
        ge=0,
        le=1,
        description='Value between 0 and 1 for fraction of vegetation coverage on '
        'Element.'
    )

    t_init: float = Field(
        ...,
        ge=0,
        description='Initial temperature of Element [K].'
    )

    horizontal: bool = Field(
        ...,
        description='Boolean value indicating if Element is horizontal or not.'
    )

    name: str = Field(
        ...,
        name='Text string for name of the Element.'
    )

    @root_validator
    def check_length(cls, values):
        """Ensure material and thickness list lengths."""
        thickness_lst = values.get('layer_thickness_lst')
        material_lst = values.get('material_lst')
        assert len(thickness_lst) == len(material_lst), 'The material_lst must ' \
            'have the same length as the layer_thickness_lst. Got lengths {} and {}, ' \
            'respectively.'
        return values

    @validator('layer_thickness_lst')
    def check_layer_thickness_lst(cls, values):
        """Ensure every list value is greater than 0."""
        assert all(v > 0 for v in values), \
            'Every value in layer_thickness_lst must be greater than 0.'
        return values

    @validator('material_lst')
    def check_material_lst(cls, values):
        """Ensure every list item is a Material object."""
        assert all(isinstance(v, Material) for v in values), \
            'Every item in material_lst must be a Material object.'
        return values


class CondType(str, Enum):
    """Cooling condensation system type."""
    air = 'AIR'
    water = 'WATER'


class Building(NoExtraBaseModel):
    """Building object specifies building characteristics."""

    type: constr(regex='^Building$') = 'Building'

    floor_height: float = Field(
        ...,
        ge=0,
        description='Floor height in meters.'
    )

    int_heat_night: float = Field(
        1,
        ge=0,
        description='Nighttime internal sensible heat gain [W/m2].'
    )

    int_heat_day: float = Field(
        1,
        ge=0,
        description='Daytime internal sensible heat gain [W/m2].'
    )

    int_heat_frad: float = Field(
        0.1,
        ge=0,
        le=1,
        description='Value between 0 and 1 for radiant fraction of internal gains.'
    )

    int_heat_flat: float = Field(
        0.1,
        ge=0,
        le=1,
        description='Value between 0 and 1 for latent fraction of internal gains.'
    )

    infil: float = Field(
        ...,
        ge=0,
        description='Infiltration rate (ACH).'
    )

    vent: float = Field(
        ...,
        ge=0,
        description='Ventilation rate (ACH).'
    )

    glazing_ratio: float = Field(
        ...,
        ge=0,
        le=1,
        description='Value between 0 and 1 for glazing ratio.'
    )

    u_value: float = Field(
        ...,
        gt=0,
        description='Window U-value including film coefficent [W/(m2-K)].'
    )

    shgc: float = Field(
        ...,
        ge=0,
        le=1,
        description='Value between 0 and 1 for window Solar Heat Gain Coefficient (SHGC).'
    )

    condtype: CondType = Field(
        'AIR',
        description='Text string for cooling condensation system type. Choose '
        'from: AIR or WATER.'
    )

    cop: float = Field(
        ...,
        ge=0,
        description='COP of cooling system (nominal).'
    )

    coolcap: float = Field(
        ...,
        ge=0,
        description='Rated cooling system capacity [W/m2].'
    )

    heateff: float = Field(
        ...,
        ge=0,
        le=1,
        description='Heating system capacity.'
    )

    initial_temp: float = Field(
        291,
        ge=0,
        description='Initial indoor air temperature [K].'
    )


class BEMDef(NoExtraBaseModel):
    """Building Energy Model (BEM) definition."""
    type: constr(regex='^BEMDef$') = 'BEMDef'

    bldtype: str = Field(
        ...,
        description='Text referring to a building type. This can be a unique identifier '
        'for a custom reference building, or the identifier of the 16 predefined '
        'DOE reference building included in the UWG. This value along with the builtera'
        'must exactly match the corresponding identifiers in the bld array in order '
        'to specify the fraction of total built stock the building occupies in the UWG '
        'simulation. To reference (or overwrite) a DOE reference building, text must '
        'be one of the following: "fullservicerestaurant", "hospital", "largehotel", '
        '"largeoffice", "mediumoffice", "midriseapartment", "outpatient", '
        '"primaryschool", "quickservicerestaurant", "secondaryschool", "smallhotel", '
        '"smalloffice", "standaloneretail", "stripmall", "supermarket", or "warehouse".'
    )

    builtera: str = Field(
        ...,
        description='Text defining building built era. Must be one of the following:'
        '"pre80" (pre-1980s), "pst80" (post-1980s), or "new" (new constrution).'
        'This value along with the bldtype must exactly match the identifiers in '
        'the bld array in order to specify the fraction of total built stock the '
        'building occupies in the UWG simulation.'
    )

    @validator('builtera')
    def check_builtera(cls, value):
        assert value in REF_BUILTERA_SET, \
            'The builtera must be one of {}.Got: {}.'.format(
                REF_BUILTERA, value.lower())

    building: Building = Field(
        ...,
        description='Building object.'
    )

    mass: Element = Field(
        ...,
        description='Element object for building internal mass.'
    )

    wall: Element = Field(
        ...,
        description='Element object for building wall.'
    )

    roof: Element = Field(
        ...,
        description='Element object for building roof.'
    )


class SchDef(NoExtraBaseModel):
    """Schedule definition class."""

    type: constr(regex='^SchDef$') = 'SchDef'

    bldtype: str = Field(
        ...,
        description='Text referring to a building type. This can be a unique identifier '
        'for a custom reference building, or the identifier of the 16 predefined '
        'DOE reference building included in the UWG. This value along with the builtera'
        'must exactly match the corresponding identifiers in the bld array in order '
        'to specify the fraction of total built stock the building occupies in the UWG '
        'simulation. To reference (or overwrite) a DOE reference building, text must '
        'be one of the following: "fullservicerestaurant", "hospital", "largehotel", '
        '"largeoffice", "mediumoffice", "midriseapartment", "outpatient", '
        '"primaryschool", "quickservicerestaurant", "secondaryschool", "smallhotel", '
        '"smalloffice", "standaloneretail", "stripmall", "supermarket", or "warehouse".'
    )

    builtera: str = Field(
        ...,
        description='Text defining building built era. Must be one of the following:'
        '"pre80" (pre-1980s), "pst80" (post-1980s), or "new" (new constrution).'
        'This value along with the bldtype must exactly match the identifiers in '
        'the bld array in order to specify the fraction of total built stock the '
        'building occupies in the UWG simulation.'
    )

    @validator('builtera')
    def check_builtera(cls, value):
        assert value in REF_BUILTERA_SET, \
            'The builtera must be one of {}.Got: {}.'.format(
                REF_BUILTERA, value.lower())

    elec: WEEK_MATRIX = Field(
        ...,
        description='Matrix of numbers for weekly electricity schedule.'
    )

    gas: WEEK_MATRIX = Field(
        default=[[0 for j in range(24)] for i in range(3)],
        description='Matrix of numbers for weekly gas schedule.'
    )

    light: WEEK_MATRIX = Field(
        ...,
        description='Matrix of numbers for weekly light schedule.'
    )

    occ: WEEK_MATRIX = Field(
        ...,
        description='Matrix of numbers for weekly occupancy schedule.'
    )

    cool: WEEK_MATRIX = Field(
        ...,
        description='Matrix of numbers for weekly cooling temperature schedule.'
    )

    heat: WEEK_MATRIX = Field(
        ...,
        description='Matrix of numbers for weekly heating temperature schedule.'
    )

    swh: WEEK_MATRIX = Field(
        default=[[0 for j in range(24)] for i in range(3)],
        description='Matrix of numbers for weekly hot water schedule.'
    )

    q_elec: float = Field(
        ...,
        ge=0,
        description='Numerical value for maximum electrical plug process load [W/m2].'
    )

    q_gas: float = Field(
        0,
        ge=0,
        description='Numerical value for maximum gas process load per unit area [W/m2].'
    )

    q_light: float = Field(
        ...,
        ge=0,
        description='Numerical value for maximum light process load per unit area '
        '[W/m2].'
    )

    n_occ: float = Field(
        ...,
        ge=0,
        description='Numerical value for maximum number of occupants per unit area '
        '[person/m2].'
    )

    vent: float = Field(
        ...,
        ge=0,
        description='Numerical value for maximum ventilation rate per unit area '
        '[m3/s/m2].'
    )

    v_swh: float = Field(
        0,
        ge=0,
        description='Numerical value for maximum hot water rate per unit area [L/hr/m2].'
    )

    @root_validator
    def check_week_matrix_values(cls, values):

        schstrlst = ['elec', 'gas', 'light', 'occ', 'cool', 'heat', 'swh']
        schlst = [values.get(schstr) for schstr in schstrlst]

        for sch in schlst:
            _sch = [hr for day in sch for hr in day]
            assert all(isinstance(v, (float, int)) for v in _sch), \
                'Every item in {} must be a number.'
        return values



from uwg_schema.model import UWG
from uwg_schema.ref_bld_template import Material, Element, Building, BEMDef, SchDef
import os

# target folder where all of the samples live
root = os.path.dirname(os.path.dirname(__file__))
target_folder = os.path.join(root, 'samples')


def test_material():
    file_path = os.path.join(target_folder, 'material.json')
    Material.parse_file(file_path)


def test_element():
    file_path = os.path.join(target_folder, 'element.json')
    Element.parse_file(file_path)


def test_building():
    file_path = os.path.join(target_folder, 'building.json')
    Building.parse_file(file_path)


def test_bemdef():
    file_path = os.path.join(target_folder, 'bemdef.json')
    BEMDef.parse_file(file_path)


def test_schdef():
    file_path = os.path.join(target_folder, 'schdef.json')
    SchDef.parse_file(file_path)


def test_uwg():
    file_path = os.path.join(target_folder, 'uwg.json')
    UWG.parse_file(file_path)


def test_custom_uwg():
    file_path = os.path.join(target_folder, 'custom_uwg.json')
    UWG.parse_file(file_path)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setuptools.setup(
    name="uwg-schema",
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    author="Ladybug Tools",
    author_email="info@ladybug.tools",
    description="Input schema for the Urban Weather Generator (UWG).",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ladybug-tools/uwg-schema",
    packages=setuptools.find_packages(exclude=["tests*"]),
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation :: CPython",
        "Operating System :: OS Independent"
    ],
)



"""UWG Model schema."""
from pydantic import Field, validator, constr, conlist
from typing import List, Union

from ._base import NoExtraBaseModel
from .ref_bld_template import BEMDef, SchDef, WEEK_MATRIX

# references
REF_ZONETYPE = ('1A', '2A', '2B', '3A', '3B-CA', '3B', '3C', '4A', '4B', '4C', '5A',
                '5B', '6A', '6B', '7', '8')
REF_ZONETYPE_SET = {'1A', '2A', '2B', '3A', '3B-CA', '3B', '3C', '4A', '4B', '4C', '5A',
                    '5B', '6A', '6B', '7', '8'}
REF_BUILTERA = ('pre80', 'pst80', 'new')
REF_BUILTERA_SET = {'pre80', 'pst80', 'new'}

# defaults
DEFAULT_BLD = [('largeoffice', 'pst80', 0.4),
               ('midriseapartment', 'pst80', 0.6)]
DEFAULT_SCHTRAFFIC = [
    [0.2, 0.2, 0.2, 0.2, 0.2, 0.4, 0.7, 0.9, 0.9, 0.6, 0.6, 0.6, 0.6, 0.6, 0.7, 0.8,
        0.9, 0.9, 0.8, 0.8, 0.7, 0.3, 0.2, 0.2],  # Weekday
    [0.2, 0.2, 0.2, 0.2, 0.2, 0.3, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.6, 0.7,
        0.7, 0.7, 0.7, 0.5, 0.4, 0.3, 0.2, 0.2],  # Saturday
    [0.2, 0.2, 0.2, 0.2, 0.2, 0.3, 0.4, 0.4, 0.4, 0.4, 0.4, 0.4, 0.4, 0.4, 0.4, 0.4,
        0.4, 0.4, 0.4, 0.4, 0.3, 0.3, 0.2, 0.2]]  # Sunday


class UWG(NoExtraBaseModel):
    """Urban Weather Generator (UWG) class."""

    type: constr(regex='^UWG$') = 'UWG'

    version: str = Field(
        default='0.0.0',
        regex=r'([0-9]+)\.([0-9]+)\.([0-9]+)',
        description='Text string for the current version of the schema.'
    )

    bldheight: float = Field(
        ...,
        ge=0,
        description='Average urban building height in meters.'
    )

    blddensity: float = Field(
        ...,
        ge=0,
        le=1,
        description='Value between 0 and 1 for building footprint density as fraction '
        'of urban area.'
    )

    vertohor: float = Field(
        ...,
        ge=0,
        description='Value for vertical-to-horizontal urban area ratio. The '
        'vertical-to-horizontal urban area ratio is calculated by dividing '
        'the urban facade area by total urban area.'
    )

    grasscover: float = Field(
        ...,
        ge=0,
        le=1,
        description='Number for fraction of urban ground covered in grass only.'
    )

    treecover: float = Field(
        ...,
        ge=0,
        le=1,
        description='Number for fraction of urban ground covered in trees.'
    )

    zone: str = Field(
        ...,
        description='Text representing an ASHRAE climate zone. This value is used '
        'to specify climate zone-specific construction, and HVAC parameters for the '
        'DOE reference building types. This will not effect the simulation if only '
        'custom reference buildings are used.  Choose from the following: "1A", "2A", '
        '"2B", "3A", "3B-CA", "3B", "3C", "4A", "4B", "4C", "5A", "5B", "6A", "6B", '
        '"7", "8".'
    )

    @validator('zone')
    def check_zone(cls, value):
        assert value in REF_ZONETYPE_SET, \
            'The zone must be one of {}.Got: {}.'.format(
                REF_ZONETYPE, value.lower())

    month: int = Field(
        1,
        ge=0,
        le=12,
        description='Number (1-12) representing simulation start month.'
    )

    day: int = Field(
        1,
        ge=1,
        le=31,
        description='Number (1-31) representing simulation start day.'
    )

    nday: int = Field(
        31,
        ge=0,
        description='Number of days to simulate.'
    )

    dtsim: int = Field(
        300,
        ge=0,
        description='Simulation time step in seconds.'
    )

    dtweather: int = Field(
        3600,
        ge=0,
        description='Number for weather data time-step in seconds.'
    )

    bld: conlist(
        conlist(
            Union[float, str], min_items=3, max_items=3),
        min_items=1) = Field(
        default=DEFAULT_BLD,
        description='List of building types, eras, and fraction of urban building '
        'stock used during simulation. This consists of a nested array, with each inner '
        'array containing a string for the building type, a string for the the built '
        'era, and a number between 0 and 1, inclusive, defining built stock fraction, '
        'i.e ("LargeOffice", "New", 0.4). The building type can refer to either one of '
        'the 16 predefined building types contained in the UWG (specifying reference '
        'models from the Department of Energy), or a custom building types. The built '
        'eras must be one of: "pre80", "pst80", or "new", referring to pre-1980s, '
        'post-1980s, or new construction. If referencing custome references, the '
        'building type, and built era referenced here must exactly match the bldtype '
        'and builtera property in the custom BEMDef and SchDef provided in the '
        'ref_bem_vector and ref_sch_vector arrays. The fractions should sum to one.'
    )

    @ validator('bld')
    def check_bld(cls, value):
        """Ensure bld arrays have correct order of types."""
        total_frac = 0.0
        for bld_row in value:
            bldtype, builtera, frac = bld_row[0], bld_row[1], bld_row[2]
            assert isinstance(bldtype, str), 'The first item in the '
            'bld array must be text defining the reference building '
            'type. Got: {}.'.format(bldtype)
            assert isinstance(builtera, str) and builtera.lower() in REF_BUILTERA_SET, \
                'The second item in the bld array must be text defining the built '
            'era as one of {}. Got: {}.'.format(
                REF_BUILTERA, builtera.lower())
            assert 0.0 <= frac <= 1.0, 'The third item in the bld array '
            'must be a value between 0 and 1, inclusive, defining the '
            'fraction of total built stock. Got: {}.'.format(frac)
            total_frac += frac

        assert abs(total_frac - 1.0) < 1e-10, 'The sum of reference building '
        'fractions defined in bld must equal one. Got: {}.'.format(
            total_frac)

        return value
    autosize: bool = Field(
        False,
        description='Boolean to set HVAC autosize.'
    )

    h_mix: float = Field(
        ...,
        ge=0,
        le=1,
        description='Value between 0 and 1 for fraction of HVAC waste heat released to '
        'street canyon. It is assumed the rest of building HVAC waste heat is released '
        'from the roof.'
    )

    sensocc: float = Field(
        100,
        ge=0,
        description='Sensible heat from occupant [W].'
    )

    latfocc: float = Field(
        0.3,
        ge=0,
        le=1,
        description='Latent heat fraction from occupant.'
    )

    radfocc: float = Field(
        0.2,
        ge=0,
        le=1,
        description='Radiant heat fraction from occupant.'
    )

    radfequip: float = Field(
        0.5,
        ge=0,
        le=1,
        description='Radiant heat fraction from equipment.'
    )

    radflight: float = Field(
        0.7,
        ge=0,
        le=1,
        description='Radiant heat fraction from electric light.'
    )

    charlength: float = Field(
        1000,
        ge=0,
        description='Value for the urban characteristic length in meters. The '
        'characteristic length is the dimension of a square that encompasses the '
        'whole neighborhood'
    )

    albroad: float = Field(
        0.1,
        ge=0,
        le=1,
        description='Value between 0 and 1 for urban road albedo.'
    )

    droad: float = Field(
        0.5,
        ge=0,
        description='Value for thickness of urban road pavement thickness in meters.'
    )

    kroad: float = Field(
        1,
        ge=0,
        le=1,
        description='Number for road pavement conductivity [W/(mK)].'
    )

    croad: float = Field(
        1600000,
        ge=0,
        description='Road pavement volumetric heat capacity [J/m^3K].'
    )

    rurvegcover: float = Field(
        0.9,
        ge=0,
        le=1,
        description='Number for fraction of rural ground covered by vegetation.'
    )

    vegstart: int = Field(
        4,
        ge=1,
        le=12,
        description='Value between 1 and 12 for the month in which vegetation starts to '
        'evapotranspire. This month corresponds to when the leaves of vegetation are '
        'assumed to be out.'
    )

    vegend: int = Field(
        10,
        ge=1,
        le=12,
        description='Value between 1 and 12 for the month in which vegetation stops '
        'evapotranspiration. This month corresponds to when the leaves of vegetation '
        'are assumed to fall.'
    )

    albveg: float = Field(
        0.25,
        ge=0,
        le=1,
        description='Number for vegetation albedo.'
    )

    latgrss: float = Field(
        0.4,
        ge=0,
        le=1,
        description='Value between 0 and 1 for fraction of latent heat absorbed by '
        'urban grass.'
    )

    lattree: float = Field(
        0.6,
        ge=0,
        le=1,
        description='Value between 0 and 1 for fraction latent heat absorbed by urban '
        'trees.'
    )

    sensanth: float = Field(
        20,
        ge=0,
        description='Value for street level anthropogenic sensible heat [W/m2]. Street '
        'level anthropogenic heat is non-building heat like heat emitted from cars, '
        'pedestrians, and street cooking.'
    )

    schtraffic: WEEK_MATRIX = Field(
        default=DEFAULT_SCHTRAFFIC,
        description='Matrix for schedule of fractional anthropogenic heat load. This '
        'property consists of a 3 x 24 matrix. Each row corresponding to a schedule '
        'for a weekday, Saturday, and Sunday, and each column corresponds to an hour in '
        'the day.'
    )

    @validator('schtraffic')
    def check_schtraffic(cls, values):
        """Ensure datatype of schtraffic values."""
        _values = [hr for day in values for hr in day]
        assert all(isinstance(v, (float, int)) for v in _values), \
            'Every item in schtraffic must be a number.'
        return values

    h_ubl1: float = Field(
        1000,
        ge=0,
        description='Daytime urban boundary layer height in meters.'
    )

    h_ubl2: float = Field(
        80,
        ge=0,
        description='Nighttime urban boundary layer height in meters.'
    )

    h_ref: float = Field(
        150,
        ge=0,
        description='Inversion height in meters.'
    )

    h_temp: float = Field(
        2,
        ge=0,
        description='Temperature measurement height in meters.'
    )

    h_wind: float = Field(
        10,
        ge=0,
        description='Wind height in meters.'
    )

    c_circ: float = Field(
        1.2,
        ge=0,
        description='Wind scaling coefficient.'
    )

    c_exch: float = Field(
        1,
        ge=0,
        description='Exchange velocity coefficient.'
    )

    maxday: float = Field(
        150,
        ge=0,
        description='Value for maximum heat flux threshold for daytime '
        'conditions [W/m2].'
    )

    maxnight: float = Field(
        20,
        ge=0,
        description='Value for maximum heat flux threshold for nighttime '
        'conditions [W/m2].'
    )

    windmin: float = Field(
        1,
        ge=0,
        description='Value for minimum wind speed in m/s.'
    )

    h_obs: float = Field(
        0.1,
        ge=0,
        description='Value for rural average obstacle height in meters.'
    )

    shgc: float = Field(
        default=None,
        ge=0,
        le=1,
        description='Value between 0 and 1 for average building glazing Solar Heat Gain '
        'Coefficient (SHGC). If value is None, a unique shgc is set for each building '
        'from the refBEM.'
    )

    albroof: float = Field(
        default=None,
        ge=0,
        le=1,
        description='Value between 0 and 1 for average building roof albedo. If value '
        'is None, a unique albroof is set for each building from the refBEM.'
    )

    glzr: float = Field(
        default=None,
        ge=0,
        le=1,
        description='Value between 0 and 1 for average building glazing ratio. If value '
        'is None, a unique glzr is set for each building from the refBEM.'
    )

    vegroof: float = Field(
        default=None,
        ge=0,
        le=1,
        description='Fraction of roofs covered in grass/shrubs. If value is None, '
        'a unique vegroof is set for each building from the refBEM.'
    )

    albwall: float = Field(
        default=None,
        ge=0,
        le=1,
        description='Value between 0 and 1 for average building albedo. If value is '
        'None, a unique albwall is set for each building from the refBEM.'
    )

    flr_h: float = Field(
        default=None,
        ge=0,
        description='Average building floor height in meters. If value is None, a '
        'unique flr_h is set for each building from the refBEM.'
    )

    ref_sch_vector: List[SchDef] = Field(
        default=None,
        description='Optional list of custom SchDef objects to override or add to '
        'the refSchedule matrix according to the SchDef bldtype and builtera values.'
        'If value is None, all BEMDef objects are referenced from the DOE typologies '
        'defined by default in the refBEM matrix.'
    )

    ref_bem_vector: List[BEMDef] = Field(
        default=None,
        description='Optional list of custom BEMDef objects to override or add to '
        'the refBEM matrix according to the BEMDef bldtype and builtera values. '
        'If value is None, all SchDef objects are referenced from the DOE typologies '
        'defined by default in the refSch matrix.'
    )



"""Base class for all objects requiring a valid names for all engines."""
from pydantic import BaseModel, Field, Extra


class NoExtraBaseModel(BaseModel):
    """Base class for all objects that are not extensible with additional keys.

    This effectively includes all objects.
    """

    class Config:
        extra = Extra.forbid
